/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 */
#include "defs.h"
#include "debug.h"
#include "prof.h"
#include "tosdefs.h"
#include <signal.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>

#define INTERRUPTS 1
#if defined(__NeXT__)
/* co-Xist has Problems with ITIMER_VIRTUAL */
#define REAL_TIMER 1
#else
#define REAL_TIMER 0
#endif

#if REAL_TIMER
#define SET_TIMER setitimer (ITIMER_REAL, &itval, &itold);
#else
#define SET_TIMER setitimer (ITIMER_VIRTUAL, &itval, &itold);
#endif

/* main.c */
extern long tv_usec;
extern int hz200pt, vblpt, audio;
static struct timeval itv;

extern B *Image, mem[];
extern void show_screen (void);
extern char *serial_dev;

static struct itimerval itval;
static struct itimerval itold;
static void (*oldsig)();
static int vblcount, systcount;

/*--------- Information copied from hardware regs at regular intervals: ------*/
volatile UL vbase;
UL old_vbase=0;
/*------------ Information directly accessed though hw regs  -----------------*/
/*                   (watch out for unaligned accesses!!!)                    */
UW *color = (UW *)MEM(M_COLORS);
/*----------------------------------------------------------------------------*/

/* This is where all hardware emulation takes place
 */
void hardware_jobs (int dummy)
{
	if (audio) audio_update();
#if 1
	if (serial_dev != NULL && serial_stat())
	{
		SM_UB(MEM(0xfffa2b),LM_UB(MEM(0xfffa2b)) | 0x80);
		flags |= F_RCV_FULL;
	}
#endif
	/* VBL */
	if (--vblcount == 0)
	{
#if INTERRUPTS
		flags |= F_VBL;
#endif
		vblcount = vblpt;
		/* maybe later? -> don't call shifter in this case? Nah... */

		/* update some variables according to hw regs */
		if (old_vbase != vbase)
		{
#if 0
			fprintf(stderr,"New screen address is %lx\n", vbase);
#endif
			old_vbase = vbase;
		}
		/* call the shifter emulation */
		/* moved to cpu.m4c */
	}
#if INTERRUPTS
	if (--systcount == 0)
	{
		systcount = hz200pt;
		if (LM_UB(MEM(0xfffa09)) & 0x20) flags |= F_200Hz;
	}
#endif
#if 0
#ifdef SYSV 
	SET_TIMER
#endif
#endif
}

void crash (int sig)
{
	fprintf (stderr, "Got a fatal signal %d [pc=%08lx]\n", sig, pc);
}

void init_hardware (void)
{
	sigset_t set;
	struct sigaction a;
	sigemptyset(&set);
	a.sa_handler = hardware_jobs;
	a.sa_mask = set;
	a.sa_flags = 0;
#if REAL_TIMER
	sigaction (SIGALRM, &a, NULL);
#else
	sigaction (SIGVTALRM, &a, NULL);
#endif
	itv.tv_sec = 0L;
	itv.tv_usec = tv_usec;
	itval.it_interval = itval.it_value = itv;
	SET_TIMER
	vblcount = vblpt;
	systcount = hz200pt;
	signal (SIGBUS, crash);
}

void kill_hardware (void)
{
#if REAL_TIMER
	setitimer (ITIMER_REAL, &itold, NULL);
	signal (SIGALRM, oldsig);
#else
	setitimer (ITIMER_VIRTUAL, &itold, NULL);
	signal (SIGVTALRM, oldsig);
#endif
}

