/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 *
 * BUGS:
 * - vro_cpyfm probabyl doesn't work properly when fancy modes are used for
 *   screen->memory copying
 * - user-defined fill patterns are global (?)
 *
 * - need to 1) debug and 2) optimize the raster ops (in that order)
 *
 * - should use own vqt_attributes() if custom screen size looks better with
 *   larger font (really? maybe I just need to patch the appropriate variables)
 */
#include "defs.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include "mem.h"
#include "missing.h"
#include "screen.h"
#include "xlib_vdi.h"

#if XBUFFER
#define XBUF(_x) _x
#else
#define XBUF(_x)
#endif

#define VDEBUG 0

/* Dual mode: must execute all attribute functions etc. in TOS too!
 *
 *
 */

#define MAX_VWK 100
#define MAX_POINTS 1000
#define MAX_VDI_COLS	256

#define FIX_COLOR(_c) (priv_cmap ? vdi_maptab[(_c)&(MAX_VDI_COLS-1)]\
					: mapcol[vdi_maptab[(_c)&(MAX_VDI_COLS-1)]])
	
#define DEV_LOCATOR 1
#define DEV_VALUATOR 2
#define DEV_CHOICE 3
#define DEV_STRING 4
#define MODE_REQUEST 1
#define MODE_SAMPLE 2

typedef struct
{
	UB height, cellheight, top, ascent, half, descent, bottom,
		width, cellwidth, point;
} sysfontinfo;

typedef struct
{
	GC gc;
	int line_color;
	int fill_color;
	int text_color;
	int line_type, line_pattern, line_width, line_ends;
	int fill_interior, fill_style, fill_perimeter;
	int wrmode;
	int font_index;
	int font_yoff;
	int font_ver, font_hor, font_eff;
	int clipping;
	XRectangle clipr;
	Pixmap fill_pattern;
	int mode[4];
	char ud_linepat[16];
	int ud_dashlen, ud_dashoff;
} VWK;

static int vdi_maptab16[]={
0,15,1,2,4,6,3,5,7,8,9,10,12,14,11,13};

static int vdi_maptab[MAX_VDI_COLS];

UL abase,fontbase;
int vdi_ok=0,linea_ok=0;
extern	int depth,vdi,fix_screen,priv_cmap;
static int vdi_planes;
static XTextItem16 text16[1000];	/* dodgy */
static VWK *vwk[MAX_VWK];
static XPoint points[MAX_POINTS];	/* later: resizable array */
static XImage *xi, *cxi, *xi_dd_1p;
static XImage *subimage;
static GC cgc,rgc;
static Pixmap tmpx, cmask, cdata,xbuffer;
static int screen;
static Visual *visual;
extern char mapcol[];
#define SYSFONTS 3
static Font sysfont[SYSFONTS];
static char *sysfontname[SYSFONTS] = {
#if 0
"-*-lucidatypewriter-medium-r-*-*-10-*-*-*-m-*-*-*",
"-*-lucidatypewriter-medium-r-*-*-12-*-*-*-m-*-*-*",
"-*-lucidatypewriter-medium-r-*-*-14-*-*-*-m-*-*-*"
#else
"6x6 system font-8-1",
"8x8 system font-9-1",
"8x16 system font-10-1"
#endif
};

sysfontinfo sfinfo[3]=
{
	{5,6,4,4,3,1,1,4,6,0},
	{8,8,6,6,4,1,1,6,8,9},
#if 0
	{15,16,13,11,8,2,2,6,8,10},
#else
	{13,16,13,11,8,2,2,7,8,10},
#endif
};

#if VDEBUG
#define T(_x) fprintf (stderr, #_x "\n")
#else
#define T(_x) 
#endif

#define IGNORE(_x) static void _x(void){T(unimpl_ ## _x);}

extern int scr_width,scr_height,scr_planes;
#define vdi_w scr_width
#define vdi_h scr_height
int vdi_done = 0;
Window xw=None;
static Window root;
extern volatile UL vbase;

#include "fill.xbm"
Pixmap *fillpat[5];
int patnum[] = {1,1,24,12,1};

/*
	TODO:

Opcode	Function

	1		Open Workstation
	2		Close Workstation
	3		Clear Workstation
	4		Update Workstation
	5		Escape Functions
		1	Inquire addressable character cells
		2	Exit alpha mode
		3	Enter alpha mode
		4	Cursor up
		5	Cursor down
		6	Cursor right
		7	Cursor left
		8	Home Cursor
		9	Erase to end of screen
	   10	Erase to end of line
	   11	Direct cursor address
	   12	Output cursor addressable text
	   15	Inquire current alpha cursor address
	   18	Place graphic cursor
	   19	Remove last graphic cursor
	6		Polyline
	7		Polymarker
	8		Text
	9		Filled area
   11		GDP
		1	Bar
		2	Arc
		3	Pie
		4	Circle
		5	Ellipse
		6	Elliptical arc
		7	Elliptical pie
		8	Rounded rectangle
		9	Filled rounded rectangle
	   10	Justified graphics text
	12		Set character height, absolute mode
	14		Set color representation
	15		Set polyline linetype
	17		Set polyline color index
	18		Set polymarker type
	20		Set polymarker color index
	21		Set text face
	22		Set graphics text color index
	23		Set fill interior style
	24		Set fill style index
	25		Set fill color index
	26		Inquire color representation
	28		Input locator
	31		Input string
	32		Set writing mode
	33		Set input mode
	35		Inquire current polyline attributes
	36		Inquire current polymarker attributes
	37		Inquire current fill area attributes
	38		Inquire current graphic text attributes
	39		Set graphic text alignment
   100		Open virtual screen workstation
   101		Close virtual screen workstation
   102		Extended inquire function
   104		Set fill perimeter visibility
   106		Set graphic text special effects
   107		Set character cell height, points mode
   108		Set polyline end styles
   109		Copy raster, opaque
   110		Transform form
   111		Set mouse form
   112		Set used-defined fill pattern
   113		Set user-defined linestyle
   114		Fill rectangle
   115		Inquire input mode
   116		Inquire text extent
   117		Inquire character cell width
   118		Exchange timer interrupt vector
   121		Copy raster, transparent
   122		Show cursor
   123		Hide cursor
   124		Sample mouse button state
   125		Exchange button change vector
   126		Exchange mouse movement vector
   127		Exchange cursor change vector
   128		Sample keyboard state information
   129		Set clipping rectangle
   130		Inquire face name and index
   131		Inquire current face information

*/

static UW work_out_buf[128];

static UL control, intin, ptsin, intout, ptsout;
#define CONTRL(_x) MEM(control+2*(_x))
#define INTIN(_x) MEM(intin+2*(_x))
#define PTSIN(_x) MEM(ptsin+2*(_x))
#define INTOUT(_x) MEM(intout+2*(_x))
#define PTSOUT(_x) MEM(ptsout+2*(_x))
#define SW(_x,_v) SM_W(_x,_v)

IGNORE(nothing)

static int get_wrmode(int mode, int fg)
{
	switch ((mode-1)&3)
	{
		case 0: return GXcopy;
		case 1: return (fg > 0 ? GXor : GXand);
		case 2: return GXxor;
		case 3: return GXandInverted;
	}
}

void vdi_redraw (int x, int y, int w, int h)
{
#if 0
	fprintf (stderr, "Redrawing (%d,%d,%d,%d)\n",x,y,w,h);
#endif
	XCopyArea (display, xbuffer, xw, rgc, x, y, w, h, x,y);
}

static void init_data (void)
{
	int i,j,k=0;
	for (i=0; i<5; i++)
	{
		fillpat[i] = malloc(sizeof(Pixmap)*patnum[i]);
		for (j=0; j<patnum[i]; j++)
			fillpat[i][j]=XCreateBitmapFromData(display,xw,&fill_bits[32*k++],
							16, 16);
	}
	for (i=0; i<SYSFONTS; i++)
		sysfont[i] = XLoadFont (display, sysfontname[i]);
}

extern Colormap cmap;

static void init_window (void)
{
	XSetWindowAttributes at;
	XEvent e;
	screen = DefaultScreen (display);
	root = RootWindow (display, screen);
	if (xw != None) XDestroyWindow (display, xw);
	visual = DefaultVisual (display, screen);
#if 0
	xw = XCreateWindow (display, root, 0, 0, vdi_w, vdi_h, 0, depth, 
						InputOutput, visual, 0, &at);
#else	
	xw = XCreateSimpleWindow (display, root, 0, 0, vdi_w, vdi_h, 0, 0, 0);
#endif
	XSelectInput (display, xw, KeyPressMask|KeyReleaseMask|ButtonPressMask \
				|ButtonReleaseMask|PointerMotionMask|ColormapChangeMask|\
				EnterWindowMask|LeaveWindowMask|ExposureMask|StructureNotifyMask);
	XStoreName (display, xw, "STonX VDI Window");
	XMapWindow (display, xw);
#if 1
	if (cmap) XSetWindowColormap (display, xw, cmap);
#endif
	XWindowEvent (display, xw, StructureNotifyMask, &e);
	XClearWindow (display, xw);
}

static void init_wk (W *w_in, W *w_out1, W *w_out2)
{
	int i;
	SW(w_out1,vdi_w-1);
	SW(w_out1+1,vdi_h-1);
	SW(w_out1+2,0);
	SW(w_out1+3,372);	/* XXX */
	SW(w_out1+4,372);
	SW(w_out1+5,3);	/* XXX */
	SW(w_out1+6,7);	/* XXX */
	SW(w_out1+7,0);
	SW(w_out1+8,6);	/* XXX */
	SW(w_out1+9,8);	/* XXX */
	SW(w_out1+10,1);	/* XXX */
	SW(w_out1+11,24);	/* XXX */
	SW(w_out1+12,12);	/* XXX */
	SW(w_out1+13,2);	/* XXX TODO -> Get default visual */
	SW(w_out1+14,10);	/* GDPs */
	for (i=1; i<=10; i++) SW(w_out1+14+i,i);
	SW(w_out1+25,3);
	SW(w_out1+26,0);
	SW(w_out1+27,3);
	SW(w_out1+28,3);
	SW(w_out1+29,3);
	SW(w_out1+30,0);
	SW(w_out1+31,3);
	SW(w_out1+32,0);
	SW(w_out1+33,3);
	SW(w_out1+34,2);
	SW(w_out1+35,0);	/* XXX Colors y/n */
	SW(w_out1+36,1);	/* XXX */
	SW(w_out1+37,1);
	SW(w_out1+38,0);	/* CELLARRAY */
	SW(w_out1+39,2);	/* XXX Mono */
	SW(w_out1+40,2);	
	SW(w_out1+41,1);	
	SW(w_out1+42,1);	
	SW(w_out1+43,1);	
	SW(w_out1+44,2);	
	SW(w_out2,5);	/* XXX min. char width */
	SW(w_out2+1,4);	/* XXX min. baseline-topline */
	SW(w_out2+2,7);	/* XXX max. char width */
	SW(w_out2+3,13);	/* XXX max. baseline-toplien */
	SW(w_out2+4,1);	/* XXX min. line width */
	SW(w_out2+5,0);	
	SW(w_out2+6,40);	/* XXX max. line width */
	SW(w_out2+7,0);	
	SW(w_out2+8,15);	/* XXX min. marker width */
	SW(w_out2+9,11);	/* XXX min. marker height */
	SW(w_out2+10,120);	/* XXX max. marker width */
	SW(w_out2+11,88);	/* XXX max. marker height */
}

#define VMASK (GCFunction|GCLineWidth|GCForeground|GCBackground|GCPlaneMask\
				|GCGraphicsExposures|GCFont)

static void make_rectangle (int x1, int y1, int x2, int y2, XRectangle *r)
{
	if (x1 < x2)
	{
		r->x = x1;
		r->width = x2-x1+1;
	}
	else
	{
		r->x = x2;
		r->width = x1-x2+1;
	}
	if (y1 < y2)
	{
		r->y = y1;
		r->height= y2-y1+1;
	}
	else
	{
		r->y = y2;
		r->height = y1-y2+1;
	}
}

/* kludge - why do I need 2 different ones? */
static void make_rectangle0 (int x1, int y1, int x2, int y2, XRectangle *r)
{
	make_rectangle(x1,y1,x2,y2,r);
	r->height--;
	r->width--;
}

static void set_clipping (VWK *v)
{
	if (v->clipping) 
		XSetClipRectangles (display, v->gc, 0, 0, &(v->clipr), 1, Unsorted);
	else
		XSetClipMask (display, v->gc, None);
}

static int oldmode;
static void done_raster (VWK *v)
{
	XSetFunction (display, v->gc, oldmode);
	set_clipping (v);
}

static void init_oraster (VWK *v, int mode)
{
	XGCValues gv;
	XGetGCValues (display, v->gc, GCFunction, &gv);
	oldmode = gv.function;
	gv.function = mode;
	gv.fill_style = FillSolid;
	gv.foreground = FIX_COLOR(1);
	gv.background = 0;
	gv.clip_mask = None;
	XChangeGC (display, v->gc, GCClipMask|GCFunction|GCFillStyle|GCForeground|GCBackground,
				&gv);
}

static void init_traster (VWK *v, int mode, int f, int b)
{
	XGCValues gv;
	XGetGCValues (display, v->gc, GCFunction, &gv);
	oldmode = gv.function;
	gv.function = mode;
	gv.fill_style = FillSolid;
	gv.foreground = FIX_COLOR(f);
	gv.background = FIX_COLOR(b);
	XChangeGC (display, v->gc, GCFunction|GCFillStyle|GCForeground|GCBackground,
				&gv);
}

static void set_font_effects (int h, int e)
{
	vwk[h]->font_eff = e;
}

static void set_fontparms (int h)
{	
	switch (vwk[h]->font_ver)
	{
		case 0:	vwk[h]->font_yoff=0; break;
		case 1:	vwk[h]->font_yoff=sfinfo[vwk[h]->font_index].half; break;
		case 2:	vwk[h]->font_yoff=sfinfo[vwk[h]->font_index].ascent; break;
		case 3:	vwk[h]->font_yoff=-sfinfo[vwk[h]->font_index].descent; break;
		case 4:	vwk[h]->font_yoff=-sfinfo[vwk[h]->font_index].bottom; break;
		case 5:	vwk[h]->font_yoff=sfinfo[vwk[h]->font_index].top; break;
	}
}

static void change_font (int h, int i)
{
	XSetFont (display, vwk[h]->gc, sysfont[i]);
	vwk[h]->font_index = i;
	set_fontparms (h);
}

static void init_font (VWK *v)
{
	XGCValues gv;
	gv.function = get_wrmode(v->wrmode, v->text_color);
	gv.foreground = FIX_COLOR(v->text_color);
	gv.background = 0; /* FIXME */
	gv.fill_style = FillSolid;
	XChangeGC (display, v->gc,
			GCBackground|GCForeground|GCFillStyle|GCFunction, &gv);
}

static void draw_text (VWK *v, int x, int y,
						XChar2b *items, int n)
{
	if (v->wrmode != 2)
	{
		XDrawImageString16 (display,xw,v->gc,x,y,items,n);
		XBUF(XDrawImageString16 (display,xbuffer,v->gc,x,y,items,n);)
	}
	else
	{
		XDrawString16 (display,xw,v->gc,x,y,items,n);
		XBUF(XDrawString16 (display,xbuffer,v->gc,x,y,items,n);)
	}
	if (v->font_eff & 1)
	{	/* dodgy */
		XDrawString16 (display,xw,v->gc,x+1,y,items,n);
		XBUF(XDrawString16 (display,xbuffer,v->gc,x+1,y,items,n);)
	}
	if (v->font_eff & 8)
	{
		int len,ly;
		len = n * sfinfo[v->font_index].cellwidth;
		ly = y+sfinfo[v->font_index].descent-1; 
		XDrawLine (display, xw, v->gc, x, ly, x+len, ly);
		XBUF (XDrawLine (display, xbuffer, v->gc, x, ly, x+len, ly);)
	}
}

static void draw_jtext (VWK *v, int x, int y,
						XTextItem16 *items, int n)
{
	XDrawText16 (display,xw,v->gc,x,y,items,n);
	XBUF(XDrawText16 (display,xbuffer,v->gc,x,y,items,n);)
	if (v->font_eff & 1)
	{	/* dodgy */
		XDrawText16 (display,xw,v->gc,x+1,y,items,n);
		XBUF(XDrawText16 (display,xbuffer,v->gc,x+1,y,items,n);)
	}
	if (v->font_eff & 8)
	{
		int len,ly;
		len = n * sfinfo[v->font_index].cellwidth;
		ly = y+sfinfo[v->font_index].descent-1; 
		XDrawLine (display, xw, v->gc, x, ly, x+len, ly);
		XBUF (XDrawLine (display, xbuffer, v->gc, x, ly, x+len, ly);)
	}
}


char user_defined_style[16];

static char *dashes[] =
{
  "", "\14\4", "\3\5\3\5", "\7\3\3\3", "\10\10", "\4\3\2\2\2\3",
  user_defined_style
};
static int dashlen[] =
{ 0, 2, 4, 4, 2, 6, 16 };
static int dashoff[] =
{ 0, 0, 0, 0, 0, 0, 0 };
static int capstyle[] =
{
  CapButt,  CapButt,  CapRound,
  CapButt,  CapButt,  CapRound,
  CapRound, CapRound, CapRound
};
	
static void init_line (VWK *v)
{
	XGCValues gv;
	int type=v->line_type;
	int dashl, dasho;
	char *dashs;

	gv.function = get_wrmode(v->wrmode, v->line_color);
	gv.foreground = FIX_COLOR(v->line_color);
	gv.background = 0;	/* FIXME */
	gv.fill_style = FillSolid;
	gv.line_style = LineSolid;
	XChangeGC (display, v->gc,
		GCBackground|GCForeground|GCFillStyle|GCLineStyle|GCFunction, &gv);

	if (type == 6)
	{
		dashl = v->ud_dashlen;
		dasho = v->ud_dashoff;
		dashs = v->ud_linepat;
	}
	else
	{
		dashl = dashlen[type];
		dasho = dashoff[type];
		dashs = dashes[type];
	}

	if(v->line_width > 1)
	{
		XSetLineAttributes(display, v->gc, v->line_width, LineSolid,
			     capstyle[v->line_ends], JoinRound);
	}
	else if (dashl==0)
	{
		XSetLineAttributes(display, v->gc, 0, LineSolid,
			     capstyle[v->line_ends], JoinRound);
	}
	else
	{
	    XSetLineAttributes(display, v->gc, 0, LineOnOffDash,
			       capstyle[v->line_ends], JoinRound);
	    XSetDashes(display, v->gc, dasho,
		       dashs, dashl);
	}
}

static void init_perimeter (VWK *v)
{
	XGCValues gv;
	int type=v->line_type;

	gv.function = get_wrmode(v->wrmode, v->fill_color);
	gv.foreground = FIX_COLOR(v->fill_color);
	gv.background = 0;	/* FIXME */
	gv.fill_style = FillSolid;
	gv.line_style = LineSolid;
	gv.line_width = 0;
	XChangeGC (display, v->gc, GCLineWidth|
		GCBackground|GCForeground|GCFillStyle|GCLineStyle|GCFunction, &gv);
}

static void init_filled (VWK *v)
{
	XGCValues gv;
#if VDEBUG || 0
	fprintf (stderr, "I: %d, S: %d, C: %d, M: %d\n", v->fill_interior,
				v->fill_style, v->fill_color, v->wrmode);
#endif
	gv.fill_style = FillOpaqueStippled;
	gv.function = get_wrmode(v->wrmode, v->fill_color);
	if (v->wrmode == 3)
	{
		gv.foreground = FIX_COLOR(1);
		gv.background = 0;
	}
	else
	{
		if (gv.function == GXand)	/* MD_TRANS with color 0 */
		{
			gv.foreground = FIX_COLOR(1);
			gv.background = 0;
		}
		else
		{
			gv.foreground = FIX_COLOR(v->fill_interior == 0?0:v->fill_color);
			gv.background = 0;
		}
	}
	if(v->fill_interior==4)
	  gv.stipple = v->fill_pattern;
	else
	  gv.stipple = fillpat[v->fill_interior][v->fill_style-1];
	XChangeGC (display, v->gc,
		GCBackground|GCForeground|GCFillStyle|GCStipple|GCFunction, &gv);
}

static void init_vwk (int i)
{
	XGCValues gv;
	if (vwk[i]==NULL) vwk[i]=malloc (sizeof(VWK));
	gv.function = GXcopy;
	gv.line_width = 1;
	gv.foreground = FIX_COLOR(1);
	gv.background = FIX_COLOR(0);
	gv.plane_mask = AllPlanes; /* 1 */
	gv.graphics_exposures = False;
	gv.font = sysfont[2];
	vwk[i]->font_index = 2;
	vwk[i]->font_yoff = 0;
	vwk[i]->font_ver = 0;
	vwk[i]->font_hor = 0;
	vwk[i]->font_eff = 0;
	vwk[i]->wrmode = 1;
	vwk[i]->gc = XCreateGC (display, xw, VMASK, &gv);
	vwk[i]->fill_color=1;
	vwk[i]->line_color=1;
	vwk[i]->text_color=1;
	vwk[i]->fill_perimeter=0;
	vwk[i]->fill_style=1;
	vwk[i]->fill_interior=0;
	vwk[i]->line_type=0;
	vwk[i]->line_pattern=0;
	vwk[i]->line_width=1;
	vwk[i]->line_ends=0;
	vwk[i]->clipping=0;
	vwk[i]->fill_pattern=XCreateBitmapFromData(display,xw,&fill_bits[32*38],
	16,16);
	bzero((char *)vwk[i]->mode,sizeof(int)*4);
	vwk[i]->ud_dashlen=16;
	vwk[i]->ud_dashoff=0;
	init_line (vwk[i]);
}

static void v_opnwk (void)
{
	XGCValues gv;
	char *sid;
	int i;
	init_window();
#if VDEBUG
	fprintf(stderr,"openvwk input:\n%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
	LM_W(INTIN(0)),
	LM_W(INTIN(1)),
	LM_W(INTIN(2)),
	LM_W(INTIN(3)),
	LM_W(INTIN(4)),
	LM_W(INTIN(5)),
	LM_W(INTIN(6)),
	LM_W(INTIN(7)),
	LM_W(INTIN(8)),
	LM_W(INTIN(9)),
	LM_W(INTIN(10)));
#endif
	init_wk (NULL, INTOUT(0), PTSOUT(0));
	init_data();
	SW(CONTRL(4),45);	/* XXX handle */
	for (i=0; i<MAX_VWK; i++)
	{
		vwk[i] = NULL;
		init_vwk (i);
	}
	xi = XCreateImage (display, visual, 1, XYBitmap, 0, NULL,
						vdi_w, vdi_h, 16, vdi_w / 8);
	xi -> bitmap_bit_order = MSBFirst;
	xi -> byte_order = MSBFirst;
	sid = malloc((vdi_w * vdi_h * depth) / 8);
#if 1
	subimage = XCreateImage (display, visual, depth,
				depth == 1 ? XYBitmap : ZPixmap, 0,  sid,
				vdi_w, vdi_h, 32, (vdi_w * depth)/8);
#else
	subimage = XCreateImage (display, visual, depth,
				depth == 1 ? XYBitmap : XYPixmap, 0,  sid,
				vdi_w, vdi_h, 32, (vdi_w * depth)/8);
#endif
	xi_dd_1p = XCreateImage (display, visual, 1, XYBitmap, 0, NULL,
							vdi_w, vdi_h, 16, vdi_w / 8);
#if XBUFFER
	xbuffer = XCreatePixmap (display, xw, vdi_w, vdi_h, depth);
	rgc = XCreateGC (display, xw, 0, NULL);
	XFillRectangle(display, xbuffer, rgc, 0, 0, vdi_w, vdi_h);
	XCopyArea (display, xw, xbuffer, rgc, 0, 0, vdi_w, vdi_h, 0, 0);
#endif
	cxi = XCreateImage (display, visual, 1, XYBitmap, 0, NULL, 16, 16, 16, 2);
	cmask = XCreatePixmap (display, xw, 16, 16, 1);
	cdata = XCreatePixmap (display, xw, 16, 16, 1);
	gv.foreground = FIX_COLOR(1);
	gv.background = FIX_COLOR(0);
	cgc = XCreateGC (display, cmask, GCForeground|GCBackground, &gv);
	tmpx = XCreatePixmap (display, xw, vdi_w, vdi_h, depth);
	vdi_done=0;
}

static void v_clrwk (void)
{
	XClearWindow (display, xw);
}

static void v_opnvwk (void)
{
	int i;
	int ph = LM_W(CONTRL(6));
	T(v_opnvwk);
	if (ph >= 11) return;
	/* do what v_opnwk() does now */
}

static void v_pline (void)
{
	int i;
	int n = LM_W(CONTRL(1));
	int h = LM_W(CONTRL(6));
	init_line (vwk[h]);
	T(v_pline);
#if VDEBUG
	fprintf(stderr,"line handle %d\n",h);
#endif
  
	for (i=0; i<n; i++)
	{
		points[i].x = LM_W(PTSIN(2*i));
		points[i].y = LM_W(PTSIN(2*i+1));
	}
#if 0
		fprintf (stderr, "Line (%d,%d) -> (%d,%d)\n", xo, yo, x, y);
#endif
	XDrawLines (display, xw, vwk[h]->gc, points, n, CoordModeOrigin);
	XBUF (XDrawLines(display, xbuffer, vwk[h]->gc,
				points, n, CoordModeOrigin);)
	vdi_done=1;
}

static void vsl_udsty (void)
{
	int i, j=0, k=-1;
  	int h = LM_W(CONTRL(6));
  	int m = LM_UW(INTIN(0));
	char *pat = vwk[h]->ud_linepat;
  	vwk[h]->line_pattern = m;
	T(vsl_udsty);
#if VDEBUG
	fprintf(stderr,"udsty handle %d\n",h);
#endif
  	/* create pattern in X format */
  
	vwk[h]->ud_dashoff=0;
  
	while((m&1) && (m != 0x0FFFF))
    {
		 vwk[h]->ud_dashoff++;
		m = (m >> 1) + 0x8000;
	}
  
	for(i=0; i<15; i++)
	{
		switch (m & 0x0C000)
		{
		case 0x8000:
			pat[j++]=i-k;
			k=i;
			m = m+m + 1;
			break;
  		case 0x4000:
			pat[j++]=i-k;
			k=i;
  		default:
			m = m+m;
 		}
	}
	if(j&1) pat[j++]=15-k;
	vwk[h]->ud_dashlen=j;
#if 0
  	  fprintf (stderr, "User defined pattern:");
  	  for(i=0; i < j; i++)
  	    fprintf (stderr, " %02x", pat[i]);
  	  fprintf (stderr, "\n");
#endif
}
  
static void vsl_type (void)
{
	int h = LM_W(CONTRL(6));
  	int m = LM_W(INTIN(0));
	T(vsl_type);
#if VDEBUG
	fprintf(stderr,"vsl_type handle %d\n",h);
#endif
  	SM_W(INTOUT(0),(vwk[h]->line_type =
  			(unsigned short)(m-1) < 7 ? (m-1) : 0)+1);
#if VDEBUG
	fprintf(stderr,"Line type: %d(%d)\n",m,vwk[h]->line_type);
#endif
}

static void vswr_mode (void)
{
	int h = LM_W(CONTRL(6));
	int m = LM_W(INTIN(0));
	T(vswr_mode);
	vwk[h]->wrmode = m;
#if VDEBUG || 0
	fprintf (stderr, "new mode: %d\n", m);
#endif
}

static void vsl_width (void)
{
	XGCValues v;
	int h = LM_W(CONTRL(6));
	int w = LM_W(PTSIN(0));
	T(vsl_width);
	vwk[h]->line_width = w;
	v.line_width = w;
#if VDEBUG
	fprintf(stderr,"Line width %d\n",w);
#endif
	XChangeGC (display, vwk[h]->gc, GCLineWidth, &v);
}

static void vsl_color (void)
{
	int h = LM_W(CONTRL(6));
	int c = LM_W(INTIN(0));
	T(vsl_color);
#if VDEBUG
	fprintf(stderr,"Line color %d\n",c);
#endif
	vwk[h]->line_color = c;
}

static void vsl_ends (void)
{
	int h = LM_W(CONTRL(6));
	int beg = LM_W(INTIN(0));
	int end = LM_W(INTIN(1));

	vwk[h]->line_ends = ((((unsigned int)beg < 3) ? beg : 0)*3 +
			     (((unsigned int)end < 3) ? end : 0));
	T(vsl_ends);
}

static void vsf_color (void)
{
	int h = LM_W(CONTRL(6));
	int c = LM_W(INTIN(0));
	T(vsf_color);
	vwk[h]->fill_color = c;
}

static void vsf_interior (void)
{
	int h = LM_W(CONTRL(6));
	int i = LM_W(INTIN(0));
	T(vsf_interior);
	SM_W(INTOUT(0),vwk[h]->fill_interior=(unsigned int)(i) < 5 ? i : 0);
	vwk[h]->fill_style=1;
}

static void vsf_style (void)
{
	int h = LM_W(CONTRL(6));
	int s = LM_W(INTIN(0));
	T(vsf_style);
	SM_W(INTOUT(0),vwk[h]->fill_style=
	     (unsigned int)(s-1) < patnum[vwk[h]->fill_interior] ? s : 1);
}

static void vsf_perimeter (void)
{
	int h = LM_W(CONTRL(6));
	int p = LM_W(INTIN(0));
	T(vsf_perimeter);
	vwk[h]->fill_perimeter=p;
}

static void vsf_udpat (void)
{
	int h = LM_W(CONTRL(6));
	char p[32];
	int i;

	for(i=0; i<16; i++)
	  {
	    p[2*i]   = LM_B(INTIN(i));
	    p[2*i+1] = LM_B(INTIN(i)+1);
	  }

	XFreePixmap(display, vwk[h]->fill_pattern);
	vwk[h]->fill_pattern = XCreateBitmapFromData(display, xw, p, 16, 16);
}


static void v_fillarea (void)
{
	int h = LM_W(CONTRL(6));
	int n = LM_W(CONTRL(1));
	int i;
	T(v_fillarea);
	for (i=0; i<n; i++)
	{
		points[i].x = LM_W(PTSIN(2*i));
		points[i].y = LM_W(PTSIN(2*i+1));
	}
	init_filled (vwk[h]);
	XFillPolygon (display, xw, vwk[h]->gc, points, n, Complex, CoordModeOrigin);
	XBUF(XFillPolygon (display, xbuffer,
		vwk[h]->gc, points, n, Complex, CoordModeOrigin);)
}

static void vr_recfl (void)
{
	XRectangle r;
	int i;
	int h = LM_W(CONTRL(6));
	T(vr_recfl);
	make_rectangle (LM_W(PTSIN(0)),LM_W(PTSIN(1)),LM_W(PTSIN(2)),LM_W(PTSIN(3)),
					&r);
	init_filled (vwk[h]);
#if VDEBUG 
	fprintf (stderr, "Rect: %d,%d +%d,%d (%d)\n", r.x,r.y,r.width,r.height,vwk[h]->fill_perimeter);
#endif
	XFillRectangle (display, xw, vwk[h]->gc, r.x, r.y, r.width, r.height);
	XBUF(XFillRectangle (display, xbuffer,
		vwk[h]->gc, r.x, r.y, r.width, r.height);)
	vdi_done=1;
}

static void vdi_gdp (void)
{
	XRectangle r;
	int i, t, s;
	int h = LM_W(CONTRL(6));
	T(GDP);
	switch (LM_W(CONTRL(5)))
	{
		case 1:	/* Bar */
			T(v_bar);
			make_rectangle (LM_W(PTSIN(0)),LM_W(PTSIN(1)),
								LM_W(PTSIN(2)),LM_W(PTSIN(3)), &r);
			init_filled (vwk[h]);
			XFillRectangle (display,xw,vwk[h]->gc, r.x, r.y, r.width, r.height);
			XBUF(XFillRectangle (display,xbuffer,
				vwk[h]->gc, r.x, r.y, r.width, r.height);)
			if (vwk[h]->fill_perimeter)
			{
				init_perimeter (vwk[h]);
				XDrawRectangle(display,xw,vwk[h]->gc,r.x,r.y,r.width,r.height);
				XBUF(XDrawRectangle(display,xbuffer,
					vwk[h]->gc,r.x,r.y,r.width,r.height);)
			}
			break;
		
		case 2:	/* Arc */
			break;

		case 3:	/* Pieslice */
			break;

		case 4:	/* Circle */
			init_filled (vwk[h]);
			t = LM_W(PTSIN(4))*2;
			XFillArc (display, xw, vwk[h]->gc,
				LM_W(PTSIN(0))-t/2, LM_W(PTSIN(1))-t/2, t, t, 0, 360*64);
			XBUF(XFillArc (display, xbuffer, vwk[h]->gc,
				LM_W(PTSIN(0))-t/2, LM_W(PTSIN(1))-t/2, t, t, 0, 360*64);)
			if (vwk[h]->fill_perimeter)
			{
				init_perimeter (vwk[h]);
				XDrawArc (display, xw, vwk[h]->gc, LM_W(PTSIN(0))-t/2,
							LM_W(PTSIN(1))-t/2, t, t, 0, 360*64);
				XBUF(XDrawArc (display, xbuffer, vwk[h]->gc, LM_W(PTSIN(0))-t/2,
							LM_W(PTSIN(1))-t/2, t, t, 0, 360*64);)
			}
			break;

		case 5:	/* Ellipse */
			init_filled (vwk[h]);
			t = LM_W(PTSIN(2));
			s = LM_W(PTSIN(3));
			XFillArc (display,xw, vwk[h]->gc, LM_W(PTSIN(0))-t,
						LM_W(PTSIN(1))-s, 2*t, 2*s, 0, 360*64);
			XBUF(XFillArc (display,xbuffer, vwk[h]->gc, LM_W(PTSIN(0))-t,
						LM_W(PTSIN(1))-s, 2*t, 2*s, 0, 360*64);)
			if (vwk[h]->fill_perimeter)
			{
				init_perimeter (vwk[h]);
				XFillArc (display,xw, vwk[h]->gc, LM_W(PTSIN(0))-t,
						LM_W(PTSIN(1))-s, 2*t, 2*s,
						0, 360*64);
				XBUF(XFillArc (display,xbuffer, vwk[h]->gc, LM_W(PTSIN(0))-t,
						LM_W(PTSIN(1))-s, 2*t, 2*s,
						0, 360*64);)
			}
			break;
						
		case 6:	/* Elliptical Arc */
			break;

		case 7:	/* Elliptical Pie */
			break;

		case 8:	/* Rounded Rectangle */
			break;

		case 9:	/* Filled Rounded Rectangle */
			break;

		case 10:/* Justified Graphics Text */
			{
				int len, w, c, x, y, n, i, delta;
				XChar2b *t;
				double s, off=0.0;
				T(v_justified);
				x = LM_W(PTSIN(0));
				y = LM_W(PTSIN(1));
				len = LM_W(PTSIN(2));
				w = LM_W(INTIN(0));
				c = LM_W(INTIN(1));
				n = LM_W(CONTRL(3))-2;
				t = (XChar2b *)INTIN(2);
#if VDEBUG
				fprintf(stderr,"Len=%d, %d chars\n", len, n);
#endif
				x -= len/2;
				len -= n*sfinfo[vwk[h]->font_index].cellwidth;
				s = (double)len/(n-1);
				delta = 0;
				init_font (vwk[h]);
				for (i=0; i<n; i++)
				{
					text16[i].chars	= &t[i];
					text16[i].nchars = 1;
					text16[i].delta = delta;
					text16[i].font = None;
					delta = ((int)(off+s))-(int)off;
					off += s;
				}
				draw_jtext (vwk[h], x, y, text16, n);
			}
			break;
	}
	vdi_done = 1;
}

static void vs_clip (void)
{
	XRectangle *r;
	int h = LM_W(CONTRL(6));
	r = &(vwk[h]->clipr);
	T(vs_clip);
	if (LM_W(INTIN(0)) == 0)
	{
#if VDEBUG
		fprintf (stderr, "Clipping off\n");
#endif
		vwk[h]->clipping = 0;
	}
	else
	{
		vwk[h]->clipping = 1;
		make_rectangle (LM_W(PTSIN(0)), LM_W(PTSIN(1)), LM_W(PTSIN(2)),
						LM_W(PTSIN(3)), r);
#if VDEBUG
		fprintf (stderr, "Clipping to (%d,%d) +(%d,%d)\n",r->x,r->y,
							r->width, r->height);
#endif
	}
	set_clipping (vwk[h]);
}

static void v_string (void)
{
#if 0
	int h = LM_W(CONTRL(6));
	int maxl = LM_W(INTIN(0));
	int echo_mode = LM_W(INTIN(1));
	int echo_x = LM_W(PTSIN(0));
	int echo_y = LM_W(PTSIN(1));
	fprintf (stderr, ">>v%s_string (%d,%d,%d) [%d,%d]\n",
		vwk[h]->mode[DEV_STRING] == 1 ? "rq":"sm",
		h, maxl, echo_mode, echo_x, echo_y);
#endif
}

static void vq_mouse (void)
{
	int dummy, x, y;
	unsigned int mask;
	Window dwin, c;
	W s=0;
	int h = LM_W(CONTRL(6));
	XQueryPointer (display, xw, &dwin, &c, &dummy, &dummy, &x, &y, &mask);
	if (mask & Button1Mask) s |= 1;
	if (mask & Button2Mask) s |= 2;
#if 0
	fprintf (stderr, "vq_mouse () -> %d,%d,%d\n", x, y, s);
#endif
	SM_W(INTOUT(0),s);
	SM_W(PTSOUT(0),x);
	SM_W(PTSOUT(1),y);
	vdi_done=1;
}

static void vq_key_s (void)
{
}

static void vsin_mode (void)
{
	int h = LM_W(CONTRL(6));
	int d = LM_W(INTIN(0));
	int m = LM_W(INTIN(1));
	vwk[h]->mode[(d-1)&3] = m;
#if VDEBUG
	if (m != 2 && d != 4)
		fprintf (stderr, "** mode for device %d set to %d\n", d, m);
#endif
}

static void v_locator (void)
{
	int h = LM_W(CONTRL(6));
#if VDEBUG
	fprintf (stderr, ">>v%s_string (%d,%d,%d)\n",
		vwk[h]->mode[DEV_LOCATOR] == 1 ? "rq":"sm", h, LM_W(PTSIN(0)),
		LM_W(PTSIN(1)));
#endif
}

/* yoff+1 is a kludge XXX */
static void v_gtext (void)
{
	int h = LM_W(CONTRL(6));
	T(v_gtext);
	init_font (vwk[h]);
#if VDEBUG
	fprintf (stderr, "v_gtext i=%d, y_off=%d\n",
		vwk[h]->font_index, vwk[h]->font_yoff);
#endif
	draw_text (vwk[h], LM_W(PTSIN(0)),
					LM_W(PTSIN(1))+vwk[h]->font_yoff+1,
					(XChar2b *)INTIN(0), LM_W(CONTRL(3)));
					
	vdi_done=1;
}

static void vst_font (void)
{
	int h = LM_W(CONTRL(6));
#if VDEBUG
	fprintf (stderr, "***** Requesting font %d!\n", LM_W(INTIN(0)));
#endif
}

static void vst_height (void)
{
	Font f;
	XFontStruct *fs;
	int i=0;
	int h = LM_W(CONTRL(6));
	int height = LM_W(PTSIN(1));
#if VEDBUG
	fprintf (stderr, "vst_height(%d)\n", height);
#endif
	do {
		if (sfinfo[i+1].top > height) break;
	} while (++i<SYSFONTS-1);
	change_font (h, i);
	SM_W(PTSOUT(0),8);
	SM_W(PTSOUT(1),sfinfo[i].height);
	SM_W(PTSOUT(2),8);
	SM_W(PTSOUT(3),sfinfo[i].cellheight);
}

static void vst_alignment (void)
{
	int h = LM_W(CONTRL(6));
	int hor = LM_W(INTIN(0));
	int ver = LM_W(INTIN(1));
#if VDEBUG
	fprintf (stderr, "vst_alignment (%d,%d)\n", hor, ver);
#endif
	SM_W(INTOUT(0),hor);
	SM_W(INTOUT(1),ver);
	vwk[h]->font_ver = ver;
	vwk[h]->font_hor = hor;
	set_fontparms (h);
}

static void vst_color (void)
{
	int h = LM_W(CONTRL(6));
	int c = LM_W(INTIN(0));
	T(vst_color);
	vwk[h]->text_color = c;
}

/* copy raster from screen to MFDB d in device dependent format */
/* doesn't work without -private at the moment FIXME */
static void copy_screen_to_st_dd (GC gc, int x, int y, int w, int h, UL s,
									UL d, int dx, int dy)
{
	UL da=LM_UL(MEM(d));
#if XBUFFER
	XCopyArea (display,xbuffer, tmpx, gc, x, y, w, h, x, y);
#else
	XCopyArea (display,xw, tmpx, gc, x, y, w, h, x, y);
#endif
	if (depth == 1 || vdi_planes == 1)
	{
#if VDEBUG 
fprintf(stderr,"copy_screen_to_dd (1/1, [%d,%d,%d,%d]->[%d,%d]\n",
			x,y,w,h,dx,dy);
#endif
		xi -> width = LM_W(MEM(d+4));
		xi -> height = LM_W(MEM(d+6));
		xi -> bytes_per_line = LM_W(MEM(d+8))*2;
		xi -> data = (char *)MEM(da);
		XGetSubImage (display, tmpx, x, y, w, h,
						1, XYPixmap, xi, dx, dy);
	}
	else	/* assume depth == 8 FIXME */
	{
		int words=LM_W(MEM(d+8));
		int height=LM_W(MEM(d+6));
		int bpp=words*2*height;
		UB *dp = MEM(da)+words*2*dy+((dx+15)&-16)/8;/*FIXME */
		UB *sp = &(subimage->data[dx+dy*vdi_w]);
		int planes=LM_W(MEM(d+12));
		int i,j,k,l;

		XGetSubImage (display, tmpx, x, y, w, h, AllPlanes,
					ZPixmap, subimage, dx, dy);
#if 0
		XPutImage (display, xw, gc, subimage, dx,dy, 320,300,w,h);
#endif
#if VDEBUG
fprintf(stderr,"copy_screen_to_dd (8/4, [%d,%d,%d,%d]->[%d,%d] MFDB[%d,%d,%d]\n",
			x,y,w,h,dx,dy,planes,words,height);
#endif
		for (i=0; i<h; i++)
		{
			for (j=0; j<words*2; j++)
			{
				for (l=0; l<planes; l++)
				{
					UB p=0;
					for (k=0; k<8; k++)
					{
						p |= ((sp[j*8+k+i*vdi_w]>>l)&1)<<(7-k);
					}
					dp[bpp*l+j+i*words*2]=p;
				}
			}
		}
	}
}

static void copy_st_dd_to_screen (GC gc, int x, int y, int w, int h, UL s,
									int dx, int dy)
{
	UL sa = LM_UL(MEM(s));
	if (depth == 1 || vdi_planes == 1)
	{
#if VDEBUG
fprintf(stderr,"copy_dd_to_screen (1/1, [%d,%d,%d,%d]->[%d,%d]\n",
			x,y,w,h,dx,dy);
#endif
		xi -> width = LM_W(MEM(s+4));
		xi -> height = LM_W(MEM(s+6));
		xi -> bytes_per_line = LM_W(MEM(s+8))*2;
		xi -> data = (char *)MEM(sa);
		XBUF(XPutImage (display,xbuffer, gc, xi, x, y, dx, dy, w, h);)
		XPutImage (display,xw, gc, xi, x, y, dx, dy, w, h);
	}
	else
	{
		UB *sp = MEM(sa)+LM_W(MEM(s+8))*2*y+((x+15)&-16)/8;/* FIXME */
		UB *dp = &(subimage->data[dx+dy*vdi_w]);
		int planes=LM_W(MEM(s+12));
		int words=LM_W(MEM(s+8));
		int height=LM_W(MEM(s+6));
		int bpp=words*2*height;
		int i,j,k,l;
#if VDEBUG
fprintf(stderr,"copy_dd_to_screen (8/%d, [%d,%d,%d,%d]->[%d,%d] MFDB[%d,%d,%d]\n",LM_W(MEM(s+12)),
			x,y,w,h,dx,dy,planes,words,height);
#endif
		for (i=0; i<h; i++)
		{
			for (j=0; j<words*2; j++)
			{
				for (k=0; k<8; k++)
				{
					UB p=0;
					for (l=0; l<planes; l++)
					{
						p |= ((sp[i*2*words+j+l*bpp] >> (7-k))&1)<<l;
					}
					dp[j*8+i*vdi_w+k]=p;
				}
			}
		}
#if XBUFFER
		XPutImage (display, xbuffer, gc, subimage, dx, dy, dx, dy, w, h);
#endif
		XPutImage (display, xw, gc, subimage, dx, dy, dx, dy, w, h);
	}
}

static void vrt_cpyfm (void)
{
	XRectangle r;
	int h = LM_W(CONTRL(6));
	UL s = LM_UL(CONTRL(7)), sa = LM_UL(MEM(s));
	UL d = LM_UL(CONTRL(9)), da = LM_UL(MEM(d));
	int sf = LM_W(MEM(s+10));
	int df = LM_W(MEM(d+10));
	int sx1,sy1,sx2,sy2;
	int dx1,dy1,dx2,dy2;
	int to_s, fr_s, c1, c2, m;
	T(vrt_cpyfm);
	sx1 = LM_W(PTSIN(0));
	sy1 = LM_W(PTSIN(1));
	sx2 = LM_W(PTSIN(2));
	sy2 = LM_W(PTSIN(3));
	dx1 = LM_W(PTSIN(4));
	dy1 = LM_W(PTSIN(5));
	dx2 = LM_W(PTSIN(6));
	dy2 = LM_W(PTSIN(7));
	c1 = LM_W(INTIN(1));
	m = get_wrmode(LM_W(INTIN(0)), c1);
	c2 = LM_W(INTIN(2));
	fr_s = (sa == 0 || sa == vbase);
	to_s = (da == 0 || da == vbase);
#if VDEBUG || 0
	fprintf (stderr, "(%d/%d %d/%d %d) S: 0x%lx (0x%lx[%d],%d,%d,%d,%d), "
					 "D: 0x%lx (0x%lx[%d], %d,%d,%d,%d)\n",fr_s,to_s,c1,c2,m,
					s, sa,sf,sx1,sy1,sx2,sy2, d, da,df,dx1,dy1,dx2,dy2);
#endif
	init_traster(vwk[h], /*twrmodes[(m-1)&3][c1>0]*/ m,c1,c2);
	if (!fr_s && to_s)
	{
		make_rectangle (sx1, sy1, sx2, sy2, &r);
		xi -> width = LM_W(MEM(s+4));
		xi -> height = LM_W(MEM(s+6));
		xi -> bytes_per_line = LM_W(MEM(s+8))*2;
		xi -> data = (char *)MEM(sa);
		XPutImage (display,xw, vwk[h]->gc, xi, r.x, r.y, dx1, dy1, 
					r.width, r.height);
		XBUF(XPutImage (display,xbuffer, vwk[h]->gc, xi, r.x, r.y, dx1, dy1, 
					r.width, r.height);)
		vdi_done=1;
	}
	done_raster(vwk[h]);
}

static void vro_cpyfm (void)
{
	XRectangle r;
	Pixmap p;
	int bpl, bpp, i, n;
	int h = LM_W(CONTRL(6));
	UL s = LM_UL(CONTRL(7)), sa = LM_UL(MEM(s));
	UL d = LM_UL(CONTRL(9)), da = LM_UL(MEM(d));
	int sf = LM_W(MEM(s+10));
	int df = LM_W(MEM(d+10));
	int sx1,sy1,sx2,sy2;
	int dx1,dy1,dx2,dy2;
	int to_s, fr_s, mode,fmt;
	T(vro_cpyfm);
	mode = LM_W(INTIN(0));
	sx1 = LM_W(PTSIN(0));
	sy1 = LM_W(PTSIN(1));
	sx2 = LM_W(PTSIN(2));
	sy2 = LM_W(PTSIN(3));
	dx1 = LM_W(PTSIN(4));
	dy1 = LM_W(PTSIN(5));
	dx2 = LM_W(PTSIN(6));
	dy2 = LM_W(PTSIN(7));
	fr_s = (sa == 0 || sa == vbase);
	to_s = (da == 0 || da == vbase);
#if VDEBUG 
	fprintf (stderr, "(%d/%d)[%d] S: 0x%lx (0x%lx[%d],%d,%d,%d,%d), "
					 "D: 0x%lx (0x%lx[%d], %d,%d,%d,%d)\n",fr_s,to_s,mode,
					s, sa,sf,sx1,sy1,sx2,sy2, d, da,df,dx1,dy1,dx2,dy2);
#endif
	if (sa == 0) sf = 0;
	if (da == 0) df = 0;
#if VDEBUG
	if (sf == 1 || df == 1)
	fprintf(stderr,"Warning: vro_cpyfm() with illegal args (rasters in standard format)\n");
#endif
	init_oraster(vwk[h],mode);
	if (fr_s && !to_s)
	{
		make_rectangle (sx1, sy1, sx2, sy2, &r);
#if 0
		bpl = LM_W(MEM(d+8))*2;
		n = LM_W(MEM(d+12));
		bpp = bpl*LM_W(MEM(d+6));
#if XBUFFER && 1
		XCopyArea (display,xbuffer, tmpx, vwk[h]->gc, r.x, r.y, r.width,
					r.height, r.x, r.y);
#else
		XCopyArea (display,xw, tmpx, vwk[h]->gc, r.x, r.y, r.width,
					r.height, r.x, r.y);
#endif
		xi_dd_1p -> width = LM_W(MEM(d+4));
		xi_dd_1p -> height = LM_W(MEM(d+6));
		xi_dd_1p -> bytes_per_line = bpl;
fprintf(stderr, "screen -> mem: %d, %d, %d, %d, %d\n", xi_dd_1p -> width, xi_dd_1p -> height,
			bpl, bpp, n);
		for (i=0; i<n; i++)
		{
			xi_dd_1p -> data = (char *)MEM(da+bpp*i);
			XGetSubImage (display, tmpx, r.x, r.y, r.width, r.height,
							1<<i, XYPixmap, xi_dd_1p, dx1, dy1);
		}
#else
		copy_screen_to_st_dd (vwk[h]->gc, r.x, r.y, r.width, r.height, s, d,
							dx1, dy1);
#endif
		vdi_done=1;
	}
	else if (!fr_s && to_s)
	{
		make_rectangle (sx1, sy1, sx2, sy2, &r);
#if 0
		bpl = LM_W(MEM(s+8))*2;
		bpp = bpl*LM_W(MEM(s+6));
		n = LM_W(MEM(s+12));
		xi_dd_1p -> width = LM_W(MEM(s+4));
		xi_dd_1p -> height = LM_W(MEM(s+6));
		xi_dd_1p -> bytes_per_line = bpl;
#if 0
		p = XCreatePixmap (display, xw, xi_dd_1p -> width,
			xi_dd_1p -> height, 1);
		xi_dd_1p -> format = XYBitmap;
#endif
fprintf(stderr, "mem -> screen: %d, %d, %d, %d, %d\n", xi_dd_1p -> width, xi_dd_1p -> height,
			bpl, bpp, n);
		for (i=0; i<n; i++)
		{
#if 0
			xi_dd_1p -> data = (char *)MEM(sa+bpp*i);
#endif
			p = XCreateBitmapFromData (display, xw, (char *)MEM(sa+bpp*i),
				 bpl*8,  LM_W(MEM(s+6)));	

#if 0
			XPutImage (display, p, vwk[h]->gc, xi_dd_1p, r.x, r.y, r.x, r.y,
					r.width, r.height);
#endif
		XSetForeground(display, vwk[h]->gc, 1<<i);
		XSetBackground(display, vwk[h]->gc, 0);
fprintf(stderr,"Copy plane (%d,%d,%d,%d) -> (%d,%d)\n",r.x,r.y,r.width,r.height,dx1,dy1);
			XBUF(XCopyPlane (display, p, xbuffer, vwk[h]->gc, r.x, r.y,
					r.width, r.height, dx1, dy1, 1);)
			XCopyPlane (display, p, xw, vwk[h]->gc, r.x, r.y,
					r.width, r.height, dx1, dy1, 1);
			XFreePixmap (display, p);
		}
#else
		copy_st_dd_to_screen (vwk[h]->gc, r.x, r.y, r.width, r.height, s, 
					dx1, dy1);
#endif
		vdi_done=1;
	}
	else if (fr_s && to_s)
	{
		make_rectangle (sx1, sy1, sx2, sy2, &r);
		XCopyArea (display, xw, xw, vwk[h]->gc, r.x, r.y, r.width, r.height,
					dx1, dy1);
#if XBUFFER
		XCopyArea (display, xbuffer, xbuffer, vwk[h]->gc, r.x, r.y, r.width, r.height,
					dx1, dy1);
#endif
		vdi_done=1;
	}
	done_raster (vwk[h]);
}

static void vsc_form (void)
{
	int h = LM_W(CONTRL(6));
	XColor cols[2]; 
	T(vsc_form);
	cxi -> byte_order = MSBFirst;
	cxi -> bitmap_bit_order = MSBFirst;
	cxi -> data = INTIN(5);
	cxi -> bitmap_pad = 32;
	XSetForeground(display, cgc, 1);
	XSetBackground(display, cgc, 0);
	XPutImage (display, cmask, cgc, cxi, 0, 0, 0, 0, 16, 16);
	cxi -> data = INTIN(21);
	XPutImage (display, cdata, cgc, cxi, 0, 0, 0, 0, 16, 16);
	cols[0].pixel = FIX_COLOR(LM_W(INTIN(4)));
	cols[1].pixel = FIX_COLOR(LM_W(INTIN(3)));
#if VDEBUG
	fprintf(stderr,"Cursor fg/bg = %d,%d\n",cols[0].pixel,cols[1].pixel);
#endif
	XQueryColors (display, cmap, cols, 2);
	XDefineCursor (display, xw, 
		XCreatePixmapCursor (display, cdata, cmask, &cols[0], &cols[1],
							LM_W(INTIN(0)), LM_W(INTIN(1))));
}

static int ax,ay,aw,ah,acw,ach;

static void vdi_escape (void)
{
	int h = LM_W(CONTRL(6));
	T(vdi_escape);
	switch(LM_W(CONTRL(5)))
	{
		case 1:
			T(vq_chcells);
			ach=16; acw=8;
			ah=vdi_h/ach; aw=vdi_w/acw;
			SM_W(INTOUT(0),ah);
			SM_W(INTOUT(1),aw);
			vdi_done=0;
			break;
		case 2:
			T(v_exit_cur);
			XClearWindow (display, xw);
			vdi_done=1;
			break;
		case 3:
			T(v_enter_cur);
			XClearWindow (display, xw);
			ax=ay=0;
			vdi_done=1;
			break;
		case 4:
			T(v_curup);
			if (ay>0) ay--;
			vdi_done=1;
			break;
		case 5:
			T(v_curdown);
			if (ay<ah-1) ay++;
			vdi_done=1;
			break;
		case 6:
			T(v_curright);
			if (ax<aw-1) ax++;
			vdi_done=1;
			break;
		case 7:
			T(v_curleft);
			if (ax>0) ax--;
			vdi_done=1;
			break;
		case 8:
			T(v_curhome);
			ax=ay=0;
			vdi_done=1;
			break;
		case 9:
			T(v_eeos);
			XClearArea(display,xw,0,ay*ah,vdi_w,vdi_h-ay*ah, False);
			vdi_done=1;
			break;
		case 10:
			T(v_eeol);
			XClearArea(display,xw,ax*aw,ay*ah,vdi_w-ax*aw,ah, False);
			vdi_done=1;
			break;
		case 11:
			T(v_curaddress);
			ax=MIN(aw-1,LM_W(INTIN(0)));
			ay=MIN(ah-1,LM_W(INTIN(1)));
			vdi_done=1;
			break;
		case 12:
			T(v_curtext);
			XDrawString16(display, xw, vwk[h]->gc, ax*aw, ay*ah,
						(XChar2b *)INTIN(0), LM_W(CONTRL(3)));
			vdi_done=1;
			break;
		case 13:
			T(v_rvon);
			vdi_done=1;
			break;
		case 14:
			T(v_rvoff);
			vdi_done=1;
			break;
		case 15:
			T(vq_curaddress);
			SM_W(INTOUT(0),ax);
			SM_W(INTOUT(1),ay);
			vdi_done=1;
			break;
		default:
			vdi_done=1;
			break;
	}
}

static void vst_point (void)
{
	int p;
	int i;
	int h = LM_W(CONTRL(6));
	T(vst_point);
	p = LM_W(INTIN(0));
#if VDEBUG
	fprintf(stderr,"Requesting point size %d\n",p);
#endif
	for (i=0; i<SYSFONTS; i++)
	{
		if (sfinfo[i].point > p) break;
	}
	i--;
	SM_W(PTSOUT(0),sfinfo[i].width);
	SM_W(PTSOUT(1),sfinfo[i].height);
	SM_W(PTSOUT(2),sfinfo[i].cellwidth);
	SM_W(PTSOUT(3),sfinfo[i].cellheight);
	SM_W(INTOUT(0),sfinfo[i].point);
	change_font (h, i);
	vdi_done=0;
}

static void vr_trnfm (void)
{
	XRectangle r;
	int width,height,i,bpl,bpp,j,n;
	int h = LM_W(CONTRL(6));
	UL s = LM_UL(CONTRL(7)), sa = LM_UL(MEM(s));
	UL d = LM_UL(CONTRL(9)), da = LM_UL(MEM(d));
	int sf = LM_W(MEM(s+10));
	int df = LM_W(MEM(d+10));
	int to_s, fr_s, mode;
	width = LM_W(MEM(s+4));
	height = LM_W(MEM(s+6));
	T(vr_trnfm);
	fr_s = (sa == 0 || sa == vbase);
	to_s = (da == 0 || da == vbase);
#if VDEBUG 
	fprintf (stderr, "(%d/%d)[%d] S: 0x%lx (0x%lx[%d]), "
					 "D: 0x%lx (0x%lx[%d]) {%d,%d,%d}\n",fr_s,to_s,LM_W(MEM(s+12)),
					s, sa,sf, d, da,df,width,height,LM_W(MEM(s+8)));
#endif
#if 0
	bpl = LM_W(MEM(s+8))*2;
	bpp = bpl*height;
	n = LM_W(MEM(s+12));
	/* FIXME for in-place */
	if (sf == 0)
	{
		for (i=0; i<height; i++)
			for (j=0; j<n; j++)
				memcpy(MEM(da)+bpl*i+j*bpp,MEM(sa)+i*(bpl+j)*4,bpl);
	}
	else
	{
		for (i=0; i<height; i++)
			for (j=0; j<n; j++)
			memcpy(MEM(da)+(j+1)*bpl*4,MEM(sa)+i*bpl+j*bpp,bpl);
	}
#else
	if (sa == 0) sa = vbase;
	if (da == 0) da = vbase;
	memcpy(MEM(da),MEM(sa),LM_W(MEM(s+8))*2*height*LM_W(MEM(s+12)));
#endif
	vdi_done=1;
}

static void vst_effects (void)
{
	int e;
	int h = LM_W(CONTRL(6));
	e = LM_W(INTIN(0));
	set_font_effects (h, e);
}

static void v_clsvwk (void)
{
	int h = LM_W(CONTRL(6));
	init_vwk(h);
}

static void v_clswk (void)
{
	int i;
	for (i=0; i<MAX_VWK; i++)
		if (vwk[i] != NULL) free(vwk[i]);
}

static void vqt_attributes (void)
{
	int h = LM_W(CONTRL(6));
	SM_W(INTOUT(0),vwk[h]->font_index);
	SM_W(INTOUT(1),vwk[h]->text_color);
	SM_W(INTOUT(2),0);
	SM_W(INTOUT(3),0);
	SM_W(INTOUT(4),vwk[h]->font_ver);
	SM_W(INTOUT(5),vwk[h]->wrmode);
	SM_W(PTSOUT(0),sfinfo[vwk[h]->font_index].width);
	SM_W(PTSOUT(1),sfinfo[vwk[h]->font_index].height);
	SM_W(PTSOUT(2),sfinfo[vwk[h]->font_index].cellwidth);
	SM_W(PTSOUT(3),sfinfo[vwk[h]->font_index].cellheight);
#if VDEBUG
	fprintf(stderr,"vqt_attributes output: (%d,%d,%d,%d,%d,%d,%d,%d,%d,%d)\n",LM_W(INTOUT(0)),LM_W(INTOUT(1)),LM_W(INTOUT(2)),LM_W(INTOUT(3)),LM_W(INTOUT(4)),LM_W(INTOUT(5)),LM_W(PTSOUT(0)),LM_W(PTSOUT(1)),LM_W(PTSOUT(2)),LM_W(PTSOUT(3)));
#endif
	vdi_done=1;
}

static void vq_extnd (void)
{
	T(vq_extnd);
	/* TODO: different physical workstations */
	memcpy(INTOUT(0),&work_out_buf[0],90);
	memcpy(PTSOUT(0),&work_out_buf[45],24);
	vdi_done=0;
}

IGNORE(vex_timv)
IGNORE(v_updwk)
IGNORE(v_pmarker)
IGNORE(v_cellarray)
IGNORE(vst_rotation)
IGNORE(vs_color)
IGNORE(vsm_type)
IGNORE(vsm_height)
IGNORE(vsm_color)
IGNORE(vq_color)
IGNORE(vq_cellaray)
IGNORE(v_valuator)	/* -"-			*/
IGNORE(v_choice)	/* -"-			*/
IGNORE(vql_attributes)
IGNORE(vqm_attributes)
IGNORE(vqf_attributes)
IGNORE(v_contourfill)
IGNORE(v_get_pixel)
IGNORE(vst_ends)
IGNORE(vqin_mode)
IGNORE(vqt_extent)
IGNORE(vqt_width)
IGNORE(vst_load_fonts)
IGNORE(vst_unload_fonts)
IGNORE(v_show_c)
IGNORE(v_hide_c)
IGNORE(vex_butv)
IGNORE(vex_motv)
IGNORE(vex_curv)
IGNORE(vqt_name)
IGNORE(vqt_fontinfo)

void (*vdifunc[])(void) = 
{
	v_opnwk,			/* X */
	v_clswk,
	v_clrwk,
	v_updwk,
	vdi_escape,
	v_pline,
	v_pmarker,
	v_gtext,
	v_fillarea,
	v_cellarray,
	vdi_gdp,
	vst_height,			/* X 5, 13 */
	vst_rotation,
	vs_color,
	vsl_type,			/* X */
	vsl_width,			/* X */
	vsl_color,			
	vsm_type,
	vsm_height,
	vsm_color,
	vst_font,
	vst_color,
	vsf_interior,
	vsf_style,
	vsf_color,
	vq_color,
	vq_cellaray,
	v_locator,	/* vrq_, vsm_ 	*/
	v_valuator,	/* -"-			*/
	v_choice,	/* -"-			*/
	v_string,	/* X  -"-			*/
	vswr_mode,
	vsin_mode,	/* X */
	nothing,
	vql_attributes,
	vqm_attributes,
	vqf_attributes,
	vqt_attributes,
	vst_alignment,
	nothing, nothing, nothing, nothing, nothing, nothing, nothing, nothing,
	nothing, nothing, nothing, nothing, nothing, nothing, nothing, nothing,
	nothing, nothing, nothing, nothing, nothing, nothing, nothing, nothing,
	nothing, nothing, nothing, nothing, nothing, nothing, nothing, nothing,
	nothing, nothing, nothing, nothing, nothing, nothing, nothing, nothing,
	nothing, nothing, nothing, nothing, nothing, nothing, nothing, nothing,
	nothing, nothing, nothing, nothing, nothing, nothing, nothing, nothing,
	nothing, nothing, nothing, nothing,
	v_opnvwk,
	v_clsvwk,
	vq_extnd,
	v_contourfill,
	vsf_perimeter,
	v_get_pixel,
	vst_effects,
	vst_point,
	vsl_ends,
	vro_cpyfm,
	vr_trnfm,		/* X */
	vsc_form,
	vsf_udpat,	
	vsl_udsty,		/* X */
	vr_recfl,
	vqin_mode,
	vqt_extent,
	vqt_width,
	vex_timv,		/* X */
	vst_load_fonts,
	vst_unload_fonts,
	vrt_cpyfm,		/* X */
	v_show_c,		/* X */
	v_hide_c,		/* X */
	vq_mouse,		/* X */
	vex_butv,		/* X */
	vex_motv,
	vex_curv,
	vq_key_s,		/* X */
	vs_clip,		/* X */
	vqt_name,
	vqt_fontinfo,
};

#define N_VDIFUNC (sizeof(vdifunc)/sizeof(vdifunc[0]))

void Vdi (UL as)
{
	UW osr = sr;
	W c;
	UL pblock;
	UNSET_Z();
	if (!vdi || DREG(0) != 115) return;
	pblock = DREG(1);
	sr |= 0x700;
	control	= LM_UL(MEM(pblock));
	intin 	= LM_UL(MEM(pblock+4));
	ptsin 	= LM_UL(MEM(pblock+8));
	intout 	= LM_UL(MEM(pblock+12));
	ptsout 	= LM_UL(MEM(pblock+16));
	c = LM_W(CONTRL(0));
#if VDEBUG
	if (c != 128 && c != 31 && c != 33)fprintf (stderr, "VDI %d: ", c);
#endif
	vdi_done=0;
	if (c >= 1 && c <= N_VDIFUNC)
		vdifunc[c-1]();
	sr = osr;
	if (vdi_done) SET_Z();
}

void linea_init (void)
{
#if 0
	if (scr_width != 640 || scr_height != 400)
	{
	fprintf (stderr,"Line-A Initialization\n");
	SM_UL(MEM(0xfa1200),LM_UL(MEM(0x28)));
	SM_UL(MEM(0x28),0xfa1204);
	}
#endif
}

void init_vdi (void)
{
#if 0
	fprintf(stderr,"Grabbing Trap #2...\n");
#endif
	SM_UL(MEM(0xfa1000),LM_UL(MEM(0x88)));
	SM_UL(MEM(0x88),0xfa1004);
#if 0
	if (fix_screen)
	{
		SM_UL(MEM(0x44e),0x300000);
		SM_UL(MEM(0x42e),0x300000);
		SM_UL(MEM(0x436),0x300000);
		vbase=0x300000;
	}
#endif
	vdi_ok=1;
}

#define CHAR_HEIGHT() (vdi && scr_height >= 400 || scr_planes==1 ? 16 : 8)

void fix_vars (void)
{
	UW *a;
	a=MEM(abase);
	SM_W(a+1,(scr_width*scr_planes)/8);
	SM_W(a-1,(scr_width*scr_planes)/8);
	SM_W(a-23,CHAR_HEIGHT());
	SM_W(a-22,scr_width/8-1);
	SM_W(a-21,scr_height/CHAR_HEIGHT()-1);
	SM_W(a-20,((scr_width*scr_planes)/8)*CHAR_HEIGHT());
	SM_W(a-6,scr_width);
	SM_W(a-2,scr_height);
#if 0
	fprintf(stderr,"FBASE=%08x, CUR_FONT=%08x, DEF_FONT=%08x, [%d,%d,%d,%d]\n",
		LM_UL(a+0x54/2),LM_UL(a-0x38a/2),LM_UL(a-0x1cc/2),
		LM_W(a-0x1f2/2),LM_W(a-0x1f0/2),LM_W(a-0x1ee/2),LM_W(a-0x1ec/2));
	if (CHAR_HEIGHT() == 16)
	{
		SM_UL(a-0x1cc/2,LM_UL(MEM(fontbase+8)));
		SM_UL(a-0x38a/2,LM_UL(MEM(fontbase+8)));
	}
	fprintf(stderr,"-> FBASE=%08x, CUR_FONT=%08x, DEF_FONT=%08x\n",
		LM_UL(a+0x54/2),LM_UL(a-0x38a/2),LM_UL(a-0x1cc/2));
#endif
}

void vdi_post (UL as)
{
	UW *a;
	UL pblock = DREG(1);
	control	= LM_UL(MEM(pblock));
	intin 	= LM_UL(MEM(pblock+4));
	ptsin 	= LM_UL(MEM(pblock+8));
	intout 	= LM_UL(MEM(pblock+12));
	ptsout 	= LM_UL(MEM(pblock+16));
	if (LM_UW(CONTRL(0)) != 1 && LM_UW(CONTRL(0)) != 100)
	{
		return;
	}
#if 0
	fprintf(stderr,"abase=$%lx (%d), screen=%lx\n",abase,fix_screen,LM_UL(MEM(0x44e)));
#endif
	a=MEM(abase);
	if (fix_screen)
	{
		SM_W(INTOUT(0),vdi_w-1);
		SM_W(INTOUT(1),vdi_h-1);
		SM_W(a-0x15a,vdi_w-1);
		SM_W(a-0x159,vdi_h-1);
#if 0
		SM_B(MEM(0x44c),3);
		fprintf(stderr,"Fixing Line-A variables...\n");
#endif
		fix_vars();
	}
	vdi_planes = LM_W(a);

	/* fix the lookup table */
	if (LM_UW(CONTRL(0)) == 1)
	{
		int i;
		for (i=0; i<MAX_VDI_COLS; i++)
			vdi_maptab[i] = (1<<vdi_planes)-1;
		for (i=0; i<(1<<vdi_planes); i++)
			vdi_maptab[i] = vdi_maptab16[i] & ((1<<vdi_planes)-1);
	}

	memcpy(&work_out_buf[0],INTOUT(0),90);
	memcpy(&work_out_buf[45],PTSOUT(0),24);
}

void linea_post (UL as)
{
	UW *a=MEM(DREG(0));
	abase=DREG(0);
	if (fix_screen) fix_vars();
}

void Init_Linea (UL as)
{
	UW *a;
	abase=DREG(0);
	fontbase=AREG(1);
	fprintf(stderr,"Initializing Line-A... (abase=%04lx)\n",
		abase);
	a = (UW *)MEM(abase);
	if (fix_screen) fix_vars();
}

